use std::{
    cell::RefCell,
    collections::HashMap,
    io::Error as IoError,
    net::SocketAddr,
    sync::{Arc, Mutex},
};

use futures_channel::mpsc::{unbounded, UnboundedSender};
use futures_util::{future, pin_mut, stream::TryStreamExt, StreamExt};

use tokio::{
    io::{AsyncRead, AsyncWrite},
    net::{TcpListener, TcpStream},
};
use tokio_tungstenite::WebSocketStream;
use tungstenite::{
    handshake::server::{ErrorResponse, Request, Response},
    protocol::Message,
};

type Tx = UnboundedSender<Message>;
type PeerMap = Arc<Mutex<HashMap<String, HashMap<SocketAddr, Tx>>>>;

async fn handle_connection(
    peer_map: PeerMap,
    ws_stream: WebSocketStream<TcpStream>,
    addr: SocketAddr,
    channel_id: String,
) {
    println!("Incomming TCP connection to channel '{}' from: {}", &channel_id, addr);

    let (tx, rx) = unbounded();
    peer_map
        .lock()
        .expect("???")
        .entry(channel_id.clone())
        .or_insert(HashMap::new())
        .insert(addr, tx);
    // peer_map.lock().unwrap().insert(addr, tx);

    let (outgoing, incomming) = ws_stream.split();

    let _channel_id = channel_id.clone();
    let broadcast_incomming = incomming.try_for_each(|msg| {
        let pm = peer_map.lock().unwrap();
        let peers = pm.get(&_channel_id).unwrap();
        if *&msg.is_ping() {
            peers.get(&addr).unwrap().unbounded_send(Message::Pong(msg.into_data())).unwrap();
            return future::ok(());
        }
        println!(
            "Received a message for channel {} from {}: {:?}",
            _channel_id,
            addr,
            msg
        );

        let broadcast_recipients = peers
        .iter()
        .filter(|(peer_addr, _)| peer_addr != &&addr)
        .map(|(_, ws_sink)| ws_sink);

        for recp in broadcast_recipients {
            recp.unbounded_send(msg.clone()).unwrap();
        }

        // We want to broadcast the message to everyone except ourselves.
        future::ok(())
    });

    let receive_from_others = rx.map(Ok).forward(outgoing);

    pin_mut!(broadcast_incomming, receive_from_others);
    future::select(broadcast_incomming, receive_from_others).await;

    println!("{} disconnected", &addr);
    let is_empty;
    {
        let mut pm = peer_map.lock().unwrap();
        let channel = pm.get_mut(&channel_id).unwrap();
        channel.remove(&addr);
        is_empty = channel.is_empty();
    }
    if is_empty {
        println!("Removing channel channel: {}", channel_id);
        peer_map.lock().unwrap().remove(&_channel_id);
    }
    // peer_map
    //     .lock()
    //     .unwrap()
    //     .get(&channel_id)
    //     .unwrap()
    //     .remove()
    //     ;

    // peer_map.lock().unwrap().
}

// https://github.com/snapview/tokio-tungstenite/issues/9#issuecomment-1316474398
async fn handle_raw_socket<T: AsyncRead + AsyncWrite + Unpin>(
    socket: T,
    channel_id: &RefCell<String>,
) -> Result<WebSocketStream<T>, tungstenite::Error> {
    let headers_callback = |req: &Request, res: Response| -> Result<Response, ErrorResponse> {
        channel_id.replace(req.uri().path().to_string().replace("/", ""));
        Ok(res)
    };
    tokio_tungstenite::accept_hdr_async(socket, headers_callback).await
}

#[tokio::main]
async fn main() -> Result<(), IoError> {
    let addr = "0.0.0.0:8080".to_string();
    let state = PeerMap::new(Mutex::new(HashMap::new()));

    let try_socket = TcpListener::bind(&addr).await;
    let listener = try_socket.expect("Failed to bind");
    println!("Listening on {}", addr);

    while let Ok((stream, addr)) = listener.accept().await {
        let channel_id = RefCell::new("".to_string());
        let ws = handle_raw_socket(stream, &channel_id)
            .await;
        let ws = match ws {
            Err(e) => {
                eprintln!("Unable to accept socket: {}", e);
                // return Ok(());
            },
            Ok(o) => {
                tokio::spawn(handle_connection(
                    state.clone(),
                    o,
                    addr,
                    channel_id.borrow().to_string(),
                ));
            },
        };

        // tokio::spawn(handle_connection(state.clone(), stream, addr, &channel_id));
    }

    Ok(())
}
